@@ -82,6 +82,7 @@ struct _HdHomeViewPrivate
 
   ClutterActor             *background;
   TidySubTexture           *background_sub;
+  MBWindowManagerClient    *live_bg;
 
   GHashTable               *applets;
 
@@ -95,6 +96,9 @@ struct _HdHomeViewPrivate
   gboolean                  move_applet_left : 1;
   gboolean                  move_applet_right : 1;
 
+  gint                      pan_gesture_start_x;
+  gint                      pan_gesture_start_y;
+
   guint                     id;
 
   guint load_background_source;
@@ -130,6 +134,8 @@ hd_home_view_allocation_changed (HdHomeView    *home_view,
                                  GParamSpec *pspec,
                                  gpointer    user_data);
 
+static void snap_widget_to_grid (ClutterActor *widget);
+
 typedef struct _HdHomeViewAppletData HdHomeViewAppletData;
 
 struct _HdHomeViewAppletData
@@ -140,6 +146,7 @@ struct _HdHomeViewAppletData
 
   guint press_cb;
   guint release_cb;
+  guint resize_cb;
   guint motion_cb;
 
   ClutterActor *close_button;
@@ -228,43 +235,10 @@ hd_home_view_class_init (HdHomeViewClass *klass)
                                                         G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 }
 
-static gboolean
-hd_home_view_background_release (ClutterActor *self,
-				 ClutterEvent *event,
-				 HdHomeView   *view)
-{
-  HdHomeViewPrivate *priv = view->priv;
-  GHashTableIter iter;
-  gpointer value;
-
-  g_debug ("Background release");
-
-  /*
-   * If the click started on an applet, we might have a motion callback
-   * installed -- remove it.
-   */
-  g_hash_table_iter_init (&iter, priv->applets);
-  while (g_hash_table_iter_next (&iter, NULL, &value))
-    {
-      HdHomeViewAppletData *data = value;
-
-      if (data->motion_cb)
-        {
-          g_signal_handler_disconnect (data->actor, data->motion_cb);
-          data->motion_cb = 0;
-        }
-    }
-
-  if (hd_render_manager_get_state() != HDRM_STATE_HOME_EDIT)
-    g_signal_emit_by_name (priv->home, "background-clicked", 0, event);
-
-  return TRUE;
-}
-
 /* applets_container is not a member of HdHomeView (it is in HdHome's 'front'
  * container. Hence we want to show/hide the applets container whenever the
  * home view itself is hidden */
-static gboolean hd_home_view_shown(HdHomeView   *view) {
+static gboolean hd_home_view_shown(HdHomeView *view) {
   HdHomeViewPrivate        *priv = view->priv;
   clutter_actor_show(priv->applets_container);
   return FALSE;
@@ -272,12 +246,98 @@ static gboolean hd_home_view_shown(HdHomeView   *view) {
 /* applets_container is not a member of HdHomeView (it is in HdHome's 'front'
  * container. Hence we want to show/hide the applets container whenever the
  * home view itself is hidden */
-static gboolean hd_home_view_hidden(HdHomeView   *view) {
+static gboolean hd_home_view_hidden(HdHomeView *view) {
   HdHomeViewPrivate        *priv = view->priv;
   clutter_actor_hide(priv->applets_container);
   return FALSE;
 }
 
+static gboolean
+is_button_press_in_gesture_start_area (ClutterEvent *event)
+{
+  g_debug ("%s. (%d, %d)",
+           __FUNCTION__,
+           event->button.x, event->button.y);
+
+  return event->button.x <= 15 || event->button.x >= HD_COMP_MGR_LANDSCAPE_WIDTH - 15;
+}
+
+static gboolean stop_pan_gesture (HdHomeView *view);
+
+static gboolean
+pan_gesture_motion (ClutterActor *actor,
+                    ClutterEvent *event,
+                    HdHomeView   *view)
+{
+  HdHomeViewPrivate *priv = view->priv;
+
+  g_debug ("%s. (%d, %d)",
+           __FUNCTION__,
+           event->motion.x, event->motion.y);
+
+  if (ABS (priv->pan_gesture_start_y - event->motion.y) > 25)
+    {
+      stop_pan_gesture (view);
+    }
+  else
+    {
+      if (ABS (priv->pan_gesture_start_x - event->motion.x) > 50)
+        {
+          stop_pan_gesture (view);
+
+          if (priv->pan_gesture_start_x <= 15)
+            hd_home_view_container_scroll_to_previous (HD_HOME_VIEW_CONTAINER (priv->view_container), 0);
+          else
+            hd_home_view_container_scroll_to_next (HD_HOME_VIEW_CONTAINER (priv->view_container), 0);
+        }
+    }
+
+  return FALSE;
+}
+
+static gboolean
+stop_pan_gesture (HdHomeView *view)
+{
+  g_debug ("%s", __FUNCTION__);
+
+  clutter_ungrab_pointer ();
+
+  g_signal_handlers_disconnect_by_func (view,
+                                        pan_gesture_motion,
+                                        view);
+  g_signal_handlers_disconnect_by_func (view,
+                                        stop_pan_gesture,
+                                        view);
+
+  return FALSE;
+}
+
+static gboolean
+pressed_on_view (ClutterActor *actor,
+                 ClutterEvent *event,
+                 HdHomeView   *view)
+{
+  HdHomeViewPrivate *priv = view->priv;
+
+  g_debug ("%s. (%d, %d)",
+           __FUNCTION__,
+           event->button.x, event->button.y);
+
+  if (is_button_press_in_gesture_start_area (event))
+    {
+      priv->pan_gesture_start_x = event->button.x;
+      priv->pan_gesture_start_y = event->button.y;
+
+      clutter_grab_pointer (CLUTTER_ACTOR (view));
+      g_signal_connect (view, "motion-event",
+                        G_CALLBACK (pan_gesture_motion), view);
+      g_signal_connect_swapped (view, "button-release-event",
+                                G_CALLBACK (stop_pan_gesture), view);
+    }
+
+  return FALSE;
+}
+
 static void
 hd_home_view_constructed (GObject *object)
 {
@@ -322,9 +382,8 @@ hd_home_view_constructed (GObject *object)
 
   clutter_actor_set_reactive (CLUTTER_ACTOR (object), TRUE);
 
-  g_signal_connect (object, "button-release-event",
-		    G_CALLBACK (hd_home_view_background_release),
-		    object);
+  g_signal_connect (object, "button-press-event",
+                    G_CALLBACK (pressed_on_view), object);
 
   g_signal_connect (object, "notify::allocation",
                     G_CALLBACK (hd_home_view_allocation_changed),
@@ -380,16 +439,86 @@ hd_home_view_finalize (GObject *object)
   G_OBJECT_CLASS (hd_home_view_parent_class)->finalize (object);
 }
 
+static void
+set_background_common (HdHomeView *hview, ClutterActor *new_bg)
+{
+  HdHomeViewPrivate *priv = hview->priv;
+  ClutterActor *actor = CLUTTER_ACTOR (hview);
+  TidySubTexture *new_bg_sub = 0;
+  ClutterColor clr = BACKGROUND_COLOR;
+
+  if (!new_bg)
+    {
+      /* Add a black background */
+      new_bg = clutter_rectangle_new_with_color (&clr);
+      clutter_actor_set_size (new_bg,
+                              HD_COMP_MGR_LANDSCAPE_WIDTH,
+                              HD_COMP_MGR_LANDSCAPE_HEIGHT);
+    }
+  else
+    {
+      guint bg_width, bg_height;
+      guint actual_width, actual_height;
+      bg_width = clutter_actor_get_width (actor);
+      bg_height = clutter_actor_get_height (actor);
+      actual_width = clutter_actor_get_width (new_bg);
+      actual_height = clutter_actor_get_height (new_bg);
+      /* It may be that we get a bigger texture than we need
+       * (because PVR texture compression has to use 2^n width
+       * and height). In this case we want to crop off the
+       * bottom + right sides, which we can do more efficiently
+       * with TidySubTexture than we can with set_clip.
+       */
+      if (bg_width != actual_width ||
+          bg_height != actual_height)
+        {
+          ClutterGeometry region;
+          region.x = 0;
+          region.y = 0;
+          region.width = actual_width > bg_width ? bg_width : actual_width;
+          region.height = actual_height > bg_height ? bg_height : actual_height;
+
+          new_bg_sub = tidy_sub_texture_new(CLUTTER_TEXTURE(new_bg));
+          tidy_sub_texture_set_region(new_bg_sub, &region);
+          clutter_actor_set_size(CLUTTER_ACTOR(new_bg_sub), bg_width, bg_height);
+          clutter_actor_hide(new_bg);
+          clutter_actor_show(CLUTTER_ACTOR(new_bg_sub));
+        }
+    }
+
+  clutter_actor_set_name (new_bg, "HdHomeView::background");
+
+  /* Add new background to the background container */
+  clutter_container_add_actor (
+              CLUTTER_CONTAINER (priv->background_container),
+              new_bg);
+  if (new_bg_sub)
+    clutter_container_add_actor (
+                CLUTTER_CONTAINER (priv->background_container),
+                CLUTTER_ACTOR(new_bg_sub));
+
+  /* Remove the old background (color or image) and the subtexture
+   * that may have been used to make it smaller */
+  if (priv->background_sub)
+      clutter_actor_destroy (CLUTTER_ACTOR(priv->background_sub));
+  if (priv->background)
+    clutter_actor_destroy (priv->background);
+
+  /* Only update blur if we're currently active */
+  if (hd_home_view_container_get_current_view(priv->view_container) == priv->id)
+    hd_render_manager_blurred_changed();
+
+  priv->background = new_bg;
+  priv->background_sub = new_bg_sub;
+}
+
 static gboolean
 load_background_idle (gpointer data)
 {
   HdHomeView *self = HD_HOME_VIEW (data);
   HdHomeViewPrivate *priv = self->priv;
-  ClutterActor *actor = CLUTTER_ACTOR (self);
   gchar *cached_background_image_file;
   ClutterActor *new_bg = 0;
-  TidySubTexture *new_bg_sub = 0;
-  ClutterColor clr = BACKGROUND_COLOR;
   GError *error = NULL;
 
   if (g_source_is_destroyed (g_main_current_source ()))
@@ -483,80 +612,81 @@ load_background_idle (gpointer data)
                  error?error->message:"");
       if (error)
         g_error_free (error);
-
-      /* Add a black background */
-      new_bg = clutter_rectangle_new_with_color (&clr);
-      clutter_actor_set_size (new_bg,
-                              HD_COMP_MGR_LANDSCAPE_WIDTH,
-                              HD_COMP_MGR_LANDSCAPE_HEIGHT);
-    }
-  else
-    {
-      guint bg_width, bg_height;
-      guint actual_width, actual_height;
-      bg_width = clutter_actor_get_width (actor);
-      bg_height = clutter_actor_get_height (actor);
-      actual_width = clutter_actor_get_width (new_bg);
-      actual_height = clutter_actor_get_height (new_bg);
-      /* It may be that we get a bigger texture than we need
-       * (because PVR texture compression has to use 2^n width
-       * and height). In this case we want to crop off the
-       * bottom + right sides, which we can do more efficiently
-       * with TidySubTexture than we can with set_clip.
-       */
-      if (bg_width != actual_width ||
-          bg_height != actual_height)
-        {
-          ClutterGeometry region;
-          region.x = 0;
-          region.y = 0;
-          region.width = actual_width > bg_width ? bg_width : actual_width;
-          region.height = actual_height > bg_height ? bg_height : actual_height;
-
-          new_bg_sub = tidy_sub_texture_new(CLUTTER_TEXTURE(new_bg));
-          tidy_sub_texture_set_region(new_bg_sub, &region);
-          clutter_actor_set_size(CLUTTER_ACTOR(new_bg_sub), bg_width, bg_height);
-          clutter_actor_hide(new_bg);
-          clutter_actor_show(CLUTTER_ACTOR(new_bg_sub));
-        }
     }
 
   g_free (cached_background_image_file);
 
-  clutter_actor_set_name (new_bg, "HdHomeView::background");
-
-  /* Add new background to the background container */
-  clutter_container_add_actor (
-              CLUTTER_CONTAINER (priv->background_container),
-              new_bg);
-  if (new_bg_sub)
-    clutter_container_add_actor (
-                CLUTTER_CONTAINER (priv->background_container),
-                CLUTTER_ACTOR(new_bg_sub));
-
-  /* Raise the texture above the solid color */
-  if (priv->background)
-    clutter_actor_raise (new_bg, priv->background);
-
-  /* Remove the old background (color or image) and the subtexture
-   * that may have been used to make it smaller */
-  if (priv->background_sub)
-      clutter_actor_destroy (CLUTTER_ACTOR(priv->background_sub));
-  if (priv->background)
-    clutter_actor_destroy (priv->background);
-
-  /* Only update blur if we're currently active */
-  if (hd_home_view_container_get_current_view (priv->view_container) == priv->id)
-    hd_render_manager_blurred_changed();
-
-  priv->background = new_bg;
-  priv->background_sub = new_bg_sub;
+  set_background_common (self, new_bg);
 
   priv->load_background_source = 0;
 
   return FALSE;
 }
 
+/* Use Window as background, mostly copied from above.
+ * 1) client != NULL means setting live-bg for this view.
+ * 2) client == NULL means unsetting the live-bg for this view. */
+void
+hd_home_view_set_live_bg (HdHomeView *view,
+                          MBWindowManagerClient *client,
+                          gboolean above_applets)
+{
+  HdHomeViewPrivate *priv = view->priv;
+  ClutterActor *new_bg = 0;
+  MBWMCompMgrClutterClient *cclient;
+
+  if (priv->load_background_source && !above_applets)
+    {
+      /* cancel ongoing background loading job unless we have transparent
+       * live background */
+      g_source_remove (priv->load_background_source);
+      priv->load_background_source = 0;
+    }
+
+  if (client) 
+    {
+      if (priv->live_bg)
+        /* remove the old one */
+        hd_home_view_set_live_bg (view, NULL, FALSE);
+
+      cclient = MB_WM_COMP_MGR_CLUTTER_CLIENT (client->cm_client);
+      new_bg = mb_wm_comp_mgr_clutter_client_get_actor (cclient);
+      clutter_actor_set_reactive (new_bg, FALSE);
+      priv->live_bg = client;
+      /* the actor is already parented to something, so reparent */
+      if (above_applets)
+        clutter_actor_reparent (new_bg, priv->applets_container);
+      else
+        clutter_actor_reparent (new_bg, priv->background_container);
+
+      if (!priv->background && above_applets)
+        /* use normal background below the live-bg */
+        hd_home_view_load_background (view);
+    }
+  else
+    {
+      /* remove it */
+      client = priv->live_bg;
+      if (client)
+        {
+          cclient = MB_WM_COMP_MGR_CLUTTER_CLIENT (client->cm_client);
+          new_bg = mb_wm_comp_mgr_clutter_client_get_actor (cclient);
+          /* it is in either container */
+          clutter_container_remove_actor (
+                CLUTTER_CONTAINER (priv->background_container), new_bg);
+          clutter_container_remove_actor (
+                CLUTTER_CONTAINER (priv->applets_container), new_bg);
+          clutter_actor_hide (new_bg);
+        }
+      priv->live_bg = NULL;
+      /* create black background to replace priv->background */
+      new_bg = priv->background = NULL;
+    }
+
+  if (!client || !above_applets)
+    set_background_common (view, new_bg);
+}
+
 void
 hd_home_view_load_background (HdHomeView *view)
 {
@@ -647,6 +777,24 @@ hd_home_view_get_view_id (HdHomeView *view)
   return priv->id;
 }
 
+static void
+hd_home_view_applet_resize (ClutterActor *applet,
+                            gpointer unused, HdHomeView *view)
+{
+  HdHomeViewAppletData *data;
+
+  if (!(data = g_hash_table_lookup (view->priv->applets, applet)))
+    return;
+  clutter_actor_set_position (data->close_button,
+                         clutter_actor_get_width (applet)
+                         - clutter_actor_get_width (data->close_button),
+                         0);
+  if (data->configure_button)
+    clutter_actor_set_position (data->configure_button, 0,
+                       clutter_actor_get_height (applet)
+                       - clutter_actor_get_height (data->close_button));
+}
+
 static gboolean
 hd_home_view_applet_motion (ClutterActor       *applet,
 			    ClutterMotionEvent *event,
@@ -687,6 +835,10 @@ hd_home_view_applet_motion (ClutterActor       *applet,
 
   /* Update applet actor position */
   clutter_actor_set_position (applet, x, y);
+  if (hd_transition_get_int ("edit_mode",
+                             "snap_to_grid_while_move",
+                             1))
+    snap_widget_to_grid (applet);
 
   /* Check if this is the only active Home view */
   if (!hd_home_view_container_get_previous_view (HD_HOME_VIEW_CONTAINER (priv->view_container)) ||
@@ -785,6 +937,38 @@ hd_home_view_applet_press (ClutterActor       *applet,
   return FALSE;
 }
 
+#define SNAP_GRID_SIZE_DEFAULT 4
+
+static gint
+snap_coordinate_to_grid (gint coordinate)
+{
+  gint snap_grid_size, offset;
+ 
+  snap_grid_size  = hd_transition_get_int ("edit_mode",
+                                           "snap_grid_size",
+                                           SNAP_GRID_SIZE_DEFAULT);
+  offset = coordinate % snap_grid_size;
+
+  if (offset > snap_grid_size / 2)
+    return coordinate - offset + snap_grid_size;
+  else
+    return coordinate - offset;
+}
+
+static void
+snap_widget_to_grid (ClutterActor *widget)
+{
+  ClutterGeometry c_geom;
+
+  /* Get applet size and position */
+  clutter_actor_get_geometry (widget, &c_geom);
+
+  c_geom.x = snap_coordinate_to_grid (c_geom.x);
+  c_geom.y = snap_coordinate_to_grid (c_geom.y);
+
+  clutter_actor_set_position (widget, c_geom.x, c_geom.y);
+}
+
 static void
 hd_home_view_store_applet_position (HdHomeView   *view,
                                     ClutterActor *applet,
@@ -798,6 +982,8 @@ hd_home_view_store_applet_position (HdHomeView   *view,
 
   data = g_hash_table_lookup (priv->applets, applet);
 
+  snap_widget_to_grid (applet);
+
   /* Get applet size and position */
   clutter_actor_get_geometry (applet, &c_geom);
 
@@ -1066,14 +1252,31 @@ close_applet (HdHomeView *view, HdHomeViewAppletData *data)
 }
 
 static gboolean
-close_button_clicked (ClutterActor       *button,
-                      ClutterButtonEvent *event,
-                      HdHomeView         *view)
+close_button_pressed (ClutterActor       *button,
+                     ClutterButtonEvent *event,
+                     HdHomeView         *view)
+{
+  HdHomeViewPrivate *priv = view->priv;
+
+  priv->applet_motion_start_x = event->x;
+  priv->applet_motion_start_y = event->y;
+
+  return TRUE;
+}
+
+static gboolean
+close_button_released (ClutterActor       *button,
+                       ClutterButtonEvent *event,
+                       HdHomeView         *view)
 {
   HdHomeViewPrivate *priv = view->priv;
   ClutterActor *applet;
   HdHomeViewAppletData *data;
 
+  if (ABS (priv->applet_motion_start_x - event->x) > MAX_TAP_DISTANCE ||
+      ABS (priv->applet_motion_start_y - event->y) > MAX_TAP_DISTANCE)
+    return TRUE;
+
   applet = clutter_actor_get_parent (button);
 
   data = g_hash_table_lookup (priv->applets, applet);
@@ -1134,15 +1337,15 @@ hd_home_view_add_applet (HdHomeView   *view,
   /* Add close button */
   close_button = hd_clutter_cache_get_texture ("AppletCloseButton.png", TRUE);
   clutter_container_add_actor (CLUTTER_CONTAINER (applet), close_button);
-  clutter_actor_set_position (close_button,
-                              clutter_actor_get_width (applet) - clutter_actor_get_width (close_button),
-                              0);
+
   clutter_actor_set_reactive (close_button, TRUE);
   clutter_actor_raise_top (close_button);
   if (!STATE_IN_EDIT_MODE (hd_render_manager_get_state ()))
     clutter_actor_hide (close_button);
   g_signal_connect (close_button, "button-press-event",
-                    G_CALLBACK (close_button_clicked), view);
+                    G_CALLBACK (close_button_pressed), view);
+  g_signal_connect (close_button, "button-release-event",
+                    G_CALLBACK (close_button_released), view);
   data->close_button = close_button;
 
   /* Add configure button */
@@ -1153,9 +1356,6 @@ hd_home_view_add_applet (HdHomeView   *view,
       configure_button = hd_clutter_cache_get_texture ("AppletConfigureButton.png", TRUE);
       clutter_container_add_actor (CLUTTER_CONTAINER (applet), configure_button);
 
-      clutter_actor_set_position (configure_button,
-                                  0,
-                                  clutter_actor_get_height (applet) - clutter_actor_get_height (close_button));
       clutter_actor_set_reactive (configure_button, TRUE);
       clutter_actor_raise_top (configure_button);
       if (!STATE_IN_EDIT_MODE (hd_render_manager_get_state ()))
@@ -1170,6 +1370,9 @@ hd_home_view_add_applet (HdHomeView   *view,
   data->press_cb = g_signal_connect (applet, "button-press-event",
                                      G_CALLBACK (hd_home_view_applet_press), view);
 
+  data->resize_cb = g_signal_connect (applet, "notify::allocation",
+                                      G_CALLBACK (hd_home_view_applet_resize), view);
+
   g_object_set_data (G_OBJECT (applet), "HD-HomeView", view);
 
   hd_home_view_load_applet_position (view,
@@ -1187,6 +1390,7 @@ hd_home_view_add_applet (HdHomeView   *view,
                                       applet,
                                       old_x,
                                       old_y);
+  hd_home_view_applet_resize (applet, NULL, view);
 
   desktop = hd_comp_mgr_get_desktop_client (HD_COMP_MGR (priv->comp_mgr));
   if (desktop)
@@ -1272,20 +1476,12 @@ hd_home_view_move_applet (HdHomeView   *view,
   mb_wm_client_stacking_mark_dirty (desktop_client);
 }
 
-ClutterActor *
-hd_home_view_get_background (HdHomeView *view)
+MBWindowManagerClient *
+hd_home_view_get_live_bg (HdHomeView *view)
 {
   HdHomeViewPrivate *priv = view->priv;
 
-  return priv->background;
-}
-
-ClutterActor *
-hd_home_view_get_applets_container (HdHomeView *view)
-{
-  HdHomeViewPrivate *priv = view->priv;
-
-  return priv->applets_container;
+  return priv->live_bg;
 }
 
 gboolean
@@ -1316,20 +1512,6 @@ hd_home_view_close_all_applets (HdHomeView *view)
     }
  }
 
-/*
-static void
-scroll_next_completed_cb (ClutterTimeline *timeline,
-                          HdHomeView      *view)
-{
-  HdHomeViewPrivate *priv = view->priv;
-
-  clutter_actor_hide (CLUTTER_ACTOR (view));
-
-  clutter_container_foreach (CLUTTER_CONTAINER (priv->applets_container),
-                             CLUTTER_CALLBACK (remove_applet_from_view),
-                             view);
- }
- */
 void
 hd_home_view_update_state (HdHomeView *view)
 {
@@ -1387,6 +1569,8 @@ applet_data_free (HdHomeViewAppletData *data)
     data->press_cb = (g_signal_handler_disconnect (data->actor, data->press_cb), 0);
   if (data->release_cb)
     data->release_cb = (g_signal_handler_disconnect (data->actor, data->release_cb), 0);
+  if (data->resize_cb)
+    data->resize_cb = (g_signal_handler_disconnect (data->actor, data->resize_cb), 0);
   if (data->motion_cb)
     data->motion_cb = (g_signal_handler_disconnect (data->actor, data->motion_cb), 0);
 
