@@ -40,6 +40,7 @@
 #include "hildon-desktop.h"
 #include "hd-launcher-grid.h"
 #include "hd-launcher-page.h"
+#include "hd-launcher-editor.h"
 #include "hd-gtk-utils.h"
 #include "hd-render-manager.h"
 #include "hd-app-mgr.h"
@@ -57,8 +58,8 @@
 
 typedef struct
 {
-  guint gsource_id;
   GList *items;
+  gboolean cancelled;
 } HdLauncherTraverseData;
 
 struct _HdLauncherPrivate
@@ -76,6 +77,9 @@ struct _HdLauncherPrivate
 
   HdLauncherTree *tree;
   HdLauncherTraverseData *current_traversal;
+
+  GtkWidget *editor;
+  gboolean editor_done;
 };
 
 #define HD_LAUNCHER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), \
@@ -104,6 +108,8 @@ static void hd_launcher_category_tile_clicked (HdLauncherTile *tile,
                                                gpointer data);
 static void hd_launcher_application_tile_clicked (HdLauncherTile *tile,
                                                   gpointer data);
+static void hd_launcher_application_tile_long_clicked (HdLauncherTile *tile,
+                                                       gpointer data);
 static gboolean hd_launcher_captured_event_cb (HdLauncher *launcher,
                                                ClutterEvent *event,
                                                gpointer data);
@@ -244,20 +250,39 @@ hd_launcher_dispose (GObject *gobject)
   G_OBJECT_CLASS (hd_launcher_parent_class)->dispose (gobject);
 }
 
+static void
+_hd_launcher_hide_page (GQuark key_id, gpointer data, gpointer user_data)
+{
+  HdLauncherPage *page = HD_LAUNCHER_PAGE (data);
+  clutter_actor_hide (CLUTTER_ACTOR(page));
+}
+
 void
 hd_launcher_show (void)
 {
   ClutterActor *self = CLUTTER_ACTOR (hd_launcher_get ());
   HdLauncherPrivate *priv = HD_LAUNCHER_GET_PRIVATE (self);
 
+  /* Make sure we hide all pages when we start,
+   * to ensure that we can't get into any bad state */
+  if (priv->pages)
+    g_datalist_foreach (&priv->pages, _hd_launcher_hide_page, NULL);
+
   ClutterActor *top_page = g_datalist_get_data (&priv->pages,
                                                 HD_LAUNCHER_ITEM_TOP_CATEGORY);
   if (!top_page)
     return;
 
   priv->active_page = top_page;
+
+  if (priv->editor_done)
+    {
+      /* TODO?: Avoid the transition if we're coming from the editor. */
+      priv->editor_done = FALSE;
+    }
+
   hd_launcher_page_transition(HD_LAUNCHER_PAGE(priv->active_page),
-        HD_LAUNCHER_PAGE_TRANSITION_IN);
+      HD_LAUNCHER_PAGE_TRANSITION_IN);
   /* We must show *after* starting the transition, because starting a new
    * transition when an old transition is in progress will cause the old
    * transition to be ended - which will in turn hide the launcher if
@@ -301,9 +326,7 @@ hd_launcher_hide_final (void)
 
 
 gboolean
-hd_launcher_back_button_clicked (ClutterActor *actor,
-                                 ClutterEvent *event,
-                                 gpointer data)
+hd_launcher_back_button_clicked ()
 {
   HdLauncherPrivate *priv = HD_LAUNCHER_GET_PRIVATE (hd_launcher_get ());
   ClutterActor *top_page = g_datalist_get_data (&priv->pages,
@@ -384,6 +407,98 @@ hd_launcher_application_tile_clicked (HdLauncherTile *tile,
                  0, data, NULL);
 }
 
+/* This is a little complex, h-d needs to:
+ * - Go back to the launcher when the user closes the editor window.
+ * - Hide and destroy the editor window when the user is taken away from it,
+ * i.e., for an incoming call.
+ */
+
+static void
+_hd_launcher_editor_is_topmost (HildonWindow *editor,
+                                GParamSpec   *pspec,
+                                HdLauncher   *launcher)
+{
+  HdLauncherPrivate *priv = launcher->priv;
+
+  g_debug ("%s: topmost: %d", __FUNCTION__,
+           hildon_window_get_is_topmost (editor));
+
+  /* If something has replaced the editor, close it. */
+  if (!hildon_window_get_is_topmost (editor))
+    {
+      gtk_widget_destroy (priv->editor);
+    }
+}
+
+static void
+_hd_launcher_editor_done (HdLauncherEditor *editor,
+                          gboolean          modified,
+                          HdLauncher       *launcher)
+{
+  HdLauncherPrivate *priv = launcher->priv;
+
+  g_debug ("%s: modified: %d", __FUNCTION__,
+           modified);
+
+  priv->editor_done = TRUE;
+
+  if (!modified)
+    {
+      /* Go back directly to the launcher. */
+      hd_render_manager_set_state (HDRM_STATE_LAUNCHER);
+    }
+}
+
+static gboolean
+_hd_launcher_editor_destroyed (GtkWidget  *widget,
+                               HdLauncher *launcher)
+{
+  HdLauncherPrivate *priv = launcher->priv;
+
+  g_debug ("%s", __FUNCTION__);
+
+  priv->editor_done = FALSE;
+  if (priv->editor)
+    {
+      priv->editor = NULL;
+    }
+
+  return FALSE;
+}
+
+static void
+hd_launcher_application_tile_long_clicked (HdLauncherTile *tile,
+                                           gpointer data)
+{
+  HdLauncher *launcher = hd_launcher_get();
+  HdLauncherPrivate *priv = launcher->priv;
+
+  priv->editor = hd_launcher_editor_new ();
+  gint x, y;
+  gfloat x_align, y_align;
+
+  clutter_actor_get_transformed_position (CLUTTER_ACTOR (tile), &x, &y);
+  x_align = (gfloat)(x + (HD_LAUNCHER_TILE_WIDTH / 2))
+                    / HD_LAUNCHER_PAGE_WIDTH;
+  y_align = (gfloat)(y + (HD_LAUNCHER_TILE_WIDTH / 2))
+                    / HD_LAUNCHER_PAGE_HEIGHT;
+
+  g_signal_connect (priv->editor, "destroy",
+                    G_CALLBACK (_hd_launcher_editor_destroyed),
+                    launcher);
+  g_signal_connect (priv->editor, "notify::is-topmost",
+                    G_CALLBACK (_hd_launcher_editor_is_topmost),
+                    launcher);
+  g_signal_connect (priv->editor, "done",
+                    G_CALLBACK (_hd_launcher_editor_done),
+                    launcher);
+
+  hd_launcher_editor_show (priv->editor);
+  hd_launcher_editor_select (HD_LAUNCHER_EDITOR (priv->editor),
+                             hd_launcher_tile_get_text (tile),
+                             x_align, y_align);
+}
+
 static void
 _hd_launcher_clear_page (GQuark key_id, gpointer data, gpointer user_data)
 {
@@ -391,17 +506,29 @@ _hd_launcher_clear_page (GQuark key_id, gpointer data, gpointer user_data)
   hd_launcher_grid_clear (HD_LAUNCHER_GRID (hd_launcher_page_get_grid (page)));
 }
 
+static void
+_hd_launcher_layout_page (GQuark key_id, gpointer data, gpointer user_data)
+{
+  HdLauncherPage *page = HD_LAUNCHER_PAGE (data);
+  hd_launcher_grid_layout(HD_LAUNCHER_GRID(hd_launcher_page_get_grid (page)));
+}
+
 static void
 hd_launcher_populate_tree_starting (HdLauncherTree *tree, gpointer data)
 {
   HdLauncher *launcher = HD_LAUNCHER (data);
   HdLauncherPrivate *priv = HD_LAUNCHER_GET_PRIVATE (launcher);
+  if (hd_render_manager_get_state () == HDRM_STATE_LAUNCHER)
+    {
+      hd_render_manager_set_state (HDRM_STATE_HOME);
+    }
+  priv->active_page = NULL;
 
   if (priv->current_traversal)
     {
-      g_source_remove(priv->current_traversal->gsource_id);
-      hd_launcher_lazy_traverse_cleanup (priv->current_traversal);
+      priv->current_traversal->cancelled = TRUE;
     }
+  priv->current_traversal = NULL;
 
   if (priv->pages)
     {
@@ -410,12 +537,6 @@ hd_launcher_populate_tree_starting (HdLauncherTree *tree, gpointer data)
       priv->pages = NULL;
     }
   g_datalist_init(&priv->pages);
-  priv->active_page = NULL;
-
-  if (hd_render_manager_get_state () == HDRM_STATE_LAUNCHER)
-    {
-      hd_render_manager_set_state (HDRM_STATE_HOME);
-    }
 }
 
 /*
@@ -446,9 +567,15 @@ hd_launcher_lazy_traverse_tree (gpointer data)
   HdLauncherTraverseData *tdata = data;
   HdLauncherItem *item;
   HdLauncherTile *tile;
-  HdLauncherPage *page;
+  HdLauncherPage *page = NULL;
   guint i;
 
+  if (!tdata ||
+      tdata->cancelled ||
+      tdata != priv->current_traversal)
+    /* This traversal is no longer current, go to cleanup. */
+    return FALSE;
+
   /* We're called back with huge latency so let's batch the work
    * to cut the overall population time. */
   for (i = 0; i < 5; i++)
@@ -460,45 +587,93 @@ hd_launcher_lazy_traverse_tree (gpointer data)
       tile = hd_launcher_tile_new (
           hd_launcher_item_get_icon_name (item),
           hd_launcher_item_get_local_name (item));
+      /* Signals can be handled during construction, so check this hasn't
+       * been cancelled. */
+      if (tdata->cancelled)
+        {
+          g_object_unref (tile);
+          return FALSE;
+        }
 
       /* Find in which page it goes */
       page = g_datalist_get_data (&priv->pages,
                                   hd_launcher_item_get_category (item));
       if (!page)
-        /* Put it in the default level. */
-        page = g_datalist_get_data (&priv->pages, HD_LAUNCHER_ITEM_DEFAULT_CATEGORY);
+        /* Put it in the top level. */
+        page = g_datalist_get_data (&priv->pages, HD_LAUNCHER_ITEM_TOP_CATEGORY);
 
-      hd_launcher_page_add_tile (page, tile);
-
-      if (hd_launcher_item_get_item_type(item) == HD_CATEGORY_LAUNCHER)
+      /* If we don't have a top level, we're in deep trouble, but we still
+       * check just in case.
+       */
+      if (!page)
         {
-          g_signal_connect (tile, "clicked",
-                            G_CALLBACK (hd_launcher_category_tile_clicked),
-                            g_datalist_get_data (&priv->pages,
-                              hd_launcher_item_get_id (item)));
+          g_warning ("%s: Couldn't find any page to accept entry %s",
+              __FUNCTION__, hd_launcher_item_get_id (item));
+          g_object_unref (tile);
         }
-      else if (hd_launcher_item_get_item_type(item) == HD_APPLICATION_LAUNCHER)
+      else
         {
-          g_signal_connect (tile, "clicked",
-                            G_CALLBACK (hd_launcher_application_tile_clicked),
-                            item);
+          hd_launcher_page_add_tile (page, tile);
+
+          if (hd_launcher_item_get_item_type(item) == HD_CATEGORY_LAUNCHER)
+            {
+              g_signal_connect (tile, "clicked",
+                                G_CALLBACK (hd_launcher_category_tile_clicked),
+                                g_datalist_get_data (&priv->pages,
+                                  hd_launcher_item_get_id (item)));
+            }
+          else if (hd_launcher_item_get_item_type(item) == HD_APPLICATION_LAUNCHER)
+            {
+              g_signal_connect (tile, "clicked",
+                                G_CALLBACK (hd_launcher_application_tile_clicked),
+                                item);
+            }
+
+          g_signal_connect (tile, "long-clicked",
+                        G_CALLBACK (hd_launcher_application_tile_long_clicked),
+                        item);
         }
 
-      g_object_unref (G_OBJECT (tdata->items->data));
+      g_object_unref (G_OBJECT (item));
       tdata->items = g_list_delete_link (tdata->items, tdata->items);
       if (!tdata->items)
-        return FALSE;
+        {
+          g_datalist_foreach(&priv->pages, _hd_launcher_layout_page, NULL);
+
+          /* This traversal has finished. */
+          priv->current_traversal = NULL;
+
+          /* If the changes came when an editor is present, switch back to
+           * launcher
+           */
+          if (priv->editor && priv->editor_done)
+            {
+              hd_render_manager_set_state (HDRM_STATE_LAUNCHER);
+            }
+          return FALSE;
+        }
     }
 
+  g_datalist_foreach(&priv->pages, _hd_launcher_layout_page, NULL);
+
   return TRUE;
 }
 
 static void
 hd_launcher_lazy_traverse_cleanup (gpointer data)
 {
-  HdLauncherPrivate *priv = HD_LAUNCHER_GET_PRIVATE (hd_launcher_get ());
+  HdLauncherTraverseData *tdata = data;
+
+  /* It's possible that the traversal has been cut short, so clean up the list. */
+  tdata->cancelled = TRUE;
+  if (tdata->items)
+    {
+      g_list_foreach (tdata->items, (GFunc)g_object_unref, NULL);
+      g_list_free (tdata->items);
+      tdata->items = NULL;
+    }
+
   g_free (data);
-  priv->current_traversal = NULL;
 }
 
 static void
@@ -516,11 +691,18 @@ hd_launcher_populate_tree_finished (HdLauncherTree *tree, gpointer data)
 
   if (priv->current_traversal)
     {
-      g_source_remove(priv->current_traversal->gsource_id);
-      hd_launcher_lazy_traverse_cleanup (priv->current_traversal);
+      priv->current_traversal->cancelled = TRUE;
     }
   priv->current_traversal = tdata;
 
+  /* if after traversal starts, the user switches to LAUNCHER,
+   * we can get empty launcher (the old page) that gets stuck on
+   * the screen until the user opens the power menu */
+  if (hd_render_manager_get_state () == HDRM_STATE_LAUNCHER)
+    {
+      hd_render_manager_set_state (HDRM_STATE_HOME);
+    }
+
   /* First we traverse the list and create all the categories,
    * so that apps can be correctly put into them.
    */
@@ -534,11 +716,10 @@ hd_launcher_populate_tree_finished (HdLauncherTree *tree, gpointer data)
   g_list_foreach (tdata->items, (GFunc) hd_launcher_create_page, NULL);
 
   /* Then we add the tiles to them in a idle callback. */
-  tdata->gsource_id = clutter_threads_add_idle_full (
-                                     CLUTTER_PRIORITY_REDRAW + 20,
-                                     hd_launcher_lazy_traverse_tree,
-                                     tdata,
-                                     hd_launcher_lazy_traverse_cleanup);
+  clutter_threads_add_idle_full (CLUTTER_PRIORITY_REDRAW + 20,
+                                 hd_launcher_lazy_traverse_tree,
+                                 tdata,
+                                 hd_launcher_lazy_traverse_cleanup);
 }
 
 /* handle clicks to the fake launch image. If we've been up this long the
@@ -600,6 +781,10 @@ hd_launcher_transition_app_start (HdLauncherApp *item)
   ClutterActor *app_image = 0;
   gint cursor_x, cursor_y;
 
+  /* Refuse to do a second loading screen if we are already showing one */
+  if (STATE_IS_LOADING(hd_render_manager_get_state()))
+      return FALSE;
+
   /* Is there a cached image? */
   if (item)
     service_name = hd_launcher_app_get_service (item);
@@ -788,6 +973,17 @@ hd_launcher_transition_app_start (HdLauncherApp *item)
   return launch_anim;
 }
 
+/* Returns whether the application starting transition is moving,
+ * ie. either the fake window or the applications stored screenshot
+ * is growing. */
+gboolean
+hd_launcher_transition_is_playing(void)
+{
+  HdLauncher *launcher = hd_launcher_get();
+  HdLauncherPrivate *priv = HD_LAUNCHER_GET_PRIVATE(launcher);
+  return clutter_timeline_is_playing(priv->launch_transition);
+}
+
 /* When a window has been created we want to be sure we've removed our
  * screenshot. Either that or we smoothly fade it out... maybe? :) */
 void hd_launcher_window_created(void)
@@ -886,7 +1082,7 @@ hd_launcher_background_clicked (HdLauncher *self,
   if (!priv->active_page ||
       (hd_launcher_page_get_drag_distance(HD_LAUNCHER_PAGE(priv->active_page)) <
                                           HD_LAUNCHER_TILE_MAX_DRAG))
-    hd_launcher_back_button_clicked(0, 0, 0);
+    hd_launcher_back_button_clicked();
 
   return TRUE;
 }
